import importlib
from datetime import datetime
from time import time

from SapphireQuant.Core.Futures import FutureAsset

from SapphireQuant.Config.TradingDay import TradingDayHelper

from SapphireQuant.Core.Enum import EnumSimulateMatchMode, EnumTradeEventType, EnumQuoteEventType
from SapphireQuant.DataStorage.Cache import FutureCache, FutureData
from SapphireQuant.DataStorage.Cache.FutureCache import EnumLiquidationType
from SapphireQuant.Simulate.Match.MatchCenter import MatchCenter
from SapphireQuant.Simulate.SimulateFutureQuoteChannel import SimulateFutureQuoteChannel
from SapphireQuant.Strategy.Function import Function


class SimulateExecutor:
    """
    回测执行
    """

    def __init__(self, **kwargs):
        self._debug_mode = False
        if 'debug_mode' in kwargs.keys():
            self._debug_mode = kwargs['debug_mode']

        self._start_simulate_date = kwargs['start_simulate_date']
        self._end_simulate_date = kwargs['end_simulate_date']
        self._start_simulate_date = TradingDayHelper.get_first_trading_day(self._start_simulate_date)
        self._end_simulate_date = TradingDayHelper.get_last_trading_day(self._end_simulate_date)

        self._drive_code = kwargs['drive_code']
        self._drive_bar_interval = kwargs['drive_bar_interval']
        self._drive_bar_type = kwargs['drive_bar_type']

        self._data_handler = kwargs['data_handler']
        self._future_codes = kwargs['future_codes']

        self._instrument_manager = self._data_handler.instrument_manager
        self.liquidation_type = EnumLiquidationType.收盘价
        if 'liquidation_type' in kwargs.keys():
            self.liquidation_type = kwargs['liquidation_type']

        self.product_code = kwargs['product_code']
        self.strategy_id = kwargs['strategy_id']
        self.fund_balance = kwargs['fund_balance']
        self.transaction_rate = kwargs['transaction_rate']
        self.future_long_margin_rate = kwargs['long_margin_rate']
        self.future_short_margin_rate = kwargs['short_margin_rate']
        self._match_mode = EnumSimulateMatchMode.order_price
        if 'match_mode' in kwargs.keys():
            self._match_mode = kwargs['match_mode']

        self._future_data = None

        self._strategy_path = kwargs['strategy_path']
        print('Load BaseStrategy:{0}'.format(self._strategy_path))
        module = importlib.import_module(kwargs['strategy_path'])
        if 'strategy_name' in kwargs.keys():
            strategy_name = kwargs['strategy_name']
        else:
            strategy_name = 'BaseStrategy'
        strategy_class = getattr(module, strategy_name)

        self._future_cache = FutureCache(self.transaction_rate, self.future_long_margin_rate, self.future_short_margin_rate,
                                         self._instrument_manager, self._data_handler, liquidation_type=self.liquidation_type)
        self._match_center = MatchCenter(self._future_cache, self._match_mode, self.future_long_margin_rate, self.future_short_margin_rate)

        kwargs['trade_handler'] = self._match_center
        kwargs['future_cache'] = self._future_cache
        self._function = Function(**kwargs)
        kwargs['function'] = self._function
        self._strategy = strategy_class(**kwargs)

        self._future_quote_channel = SimulateFutureQuoteChannel(self._drive_bar_interval, self._drive_bar_type,
                                                                self._start_simulate_date, self._end_simulate_date, self._data_handler, debug_mode=self._debug_mode)
        self._future_quote_channel.set_drive_code(self._drive_code)
        self._future_quote_channel.subscribe(self._future_codes)

        self._match_center.register_event(self.on_future_trade, EnumTradeEventType.on_trade)
        self._match_center.register_event(self.on_future_order, EnumTradeEventType.on_order)
        self._match_center.register_event(self.on_future_order_reject, EnumTradeEventType.on_order_reject)
        self._match_center.register_event(self.on_future_order_canceled, EnumTradeEventType.on_order_canceled)
        self._match_center.register_event(self.on_future_order_cancel_failed, EnumTradeEventType.on_order_cancel_failed)

        self._future_quote_channel.register_event(self.on_start, EnumQuoteEventType.on_start)
        self._future_quote_channel.register_event(self.on_day_begin, EnumQuoteEventType.on_day_begin)
        self._future_quote_channel.register_event(self.on_day_end, EnumQuoteEventType.on_day_end)
        self._future_quote_channel.register_event(self.on_tick, EnumQuoteEventType.on_tick)
        self._future_quote_channel.register_event(self.on_bar, EnumQuoteEventType.on_bar)
        self._future_quote_channel.register_event(self.on_end, EnumQuoteEventType.on_end)

        self._start_time = None

    def start(self):
        """
        开始
        """
        self._start_time = time()
        self._future_quote_channel.start()

    def stop(self):
        """
        停止
        """
        pass

    def on_start(self):
        """
        回测开始
        """
        print('#{0} | 复盘开始于={1}'.format(datetime.now().strftime('%Y/%m/%d %H:%M:%S'), self._start_simulate_date.strftime('%Y/%m/%d')))
        self._strategy.on_start()

    def on_end(self):
        """
        回测结束
        """
        self._future_cache.on_end()  # 生成相应交易信息的DataFrame
        self._function.on_end()  # 初始化Function种交易信息的DataFrame、策略的Summary的DataFrame、策略的参数DataFrame
        self._strategy.on_end()  # 回测结束，调用策略的结束函数
        print('#{0} | 复盘结束于={1}'.format(datetime.now().strftime('%Y/%m/%d %H:%M:%S'), self._end_simulate_date.strftime('%Y/%m/%d')))
        n = TradingDayHelper.get_trading_day_count(self._start_simulate_date, self._end_simulate_date)
        run_seconds = time() - self._start_time
        if n == 0:
            print('运行时间={0}秒, 平均每天={1}秒'.format(run_seconds, run_seconds))
        else:
            print('运行时间={0}秒, 平均每天={1}秒'.format(run_seconds, run_seconds / n))

    def on_day_begin(self, trading_date):
        """
        交易日开始
        :param trading_date:
        """
        begin_date = TradingDayHelper.get_first_trading_day(trading_date)
        print('#复盘{0}|开盘,交易日={1}'.format(datetime.now().strftime('%Y/%m/%d %H:%M:%S'), trading_date.strftime('%Y/%m/%d')))
        self._function.day_begin(trading_date)
        if begin_date == trading_date:
            self._future_data = FutureData()
            self._future_data.asset = FutureAsset()
            self._future_data.asset.product_code = self.product_code
            self._future_data.asset.strategy_id = self.strategy_id
            self._future_data.asset.fund_balance = self.fund_balance
            self._future_data.asset.cash = self.fund_balance
            self._future_data.asset.available_cash = self.fund_balance
            self._future_data.asset.pre_fund_balance = self.fund_balance
            self._future_data.asset.trading_date = trading_date
            self._future_data.next_asset = self._future_data.asset
        self._future_cache.day_begin(trading_date, self._future_data)
        self._strategy.on_init()

    def on_day_end(self):
        """
        交易日结束
        """
        self._future_data = self._future_cache.day_end()

    def on_tick(self, drive_data, lst_replay_data):
        """

        :param drive_data:
        :param lst_replay_data:
        """
        self._match_center.on_tick(drive_data)
        self._match_center.match_by_tick(drive_data)

        for data in lst_replay_data:
            if data is None or data.last_price <= 0:
                continue
            self._match_center.match_by_tick(data)

        for data in lst_replay_data:
            self._future_cache.on_tick(data)
            self._function.on_tick(data)

        self._future_cache.on_tick(drive_data)

        self._function.on_tick(drive_data)
        if drive_data.instrument_id == self._function.drive_code:
            self._strategy.on_tick(drive_data)

    def on_bar(self, drive_data, lst_replay_data):
        """

        :param drive_data:
        :param lst_replay_data:
        """
        for data in lst_replay_data:
            self._future_cache.on_bar(data)
            self._function.on_bar(data)

        self._future_cache.on_bar(drive_data)
        self._function.on_bar(drive_data)
        self._strategy.on_bar(drive_data)

        self._match_center.on_bar(drive_data)
        self._match_center.match_by_bar(drive_data)

        for data in lst_replay_data:
            if data is None or data.close <= 0:
                continue
            self._match_center.match_by_bar(data)

    def on_future_trade(self, trade):
        """

        :param trade:
        """
        self._future_cache.on_trade(trade)
        self._strategy.on_future_trade(trade)

    def on_future_order(self, order):
        """

        :param order:
        """
        self._future_cache.on_order(order)
        self._strategy.on_future_order(order)

    def on_future_order_reject(self, order):
        """

        :param order:
        """
        self._future_cache.on_order_reject(order)
        self._strategy.on_future_order_reject(order)

    def on_future_order_canceled(self, order):
        """

        :param order:
        """
        self._future_cache.on_order_canceled(order)
        self._strategy.on_future_order_canceled(order)

    def on_future_order_cancel_failed(self, order_guid, msg):
        """

        :param order_guid:
        :param msg:
        """
        self._future_cache.on_order_cancel_failed(order_guid, msg)
        self._strategy.on_future_order_cancel_failed(order_guid, msg)



import datetime
import string
import threading
from collections import defaultdict

from SapphireQuant import Tick
from SapphireQuant.Core import QiEvent

from SapphireQuant.Core.Bar import Bar
from SapphireQuant.Core.BarSeries import BarSeries
from SapphireQuant.Core.Enum.EnumBarType import EnumBarType
from SapphireQuant.Core.QiEvent import HandlerType
from SapphireQuant.DataProcessing.BaseBarHelper import BaseBarHelper

EVENT_BAR_OPEN = "eBarOpen."
EVENT_BAR_CLOSE = "eBarClose."


class BarProvider:
    """
    K线提供器
    """

    def __init__(self):
        self._bar_series: BarSeries = BarSeries()
        self._lst_date_time_slices: list = []
        self._pos_time: int = 0
        self._pos_bar: int = -1
        self._last_tick: Tick = None
        self._diff_time_span: datetime = datetime.timedelta(0, 0, 0)
        self._trading_date: datetime = datetime.datetime(1990, 1, 1, 0, 0, 0)
        self._is_supplement_blank_bar: bool = False
        self._enable_live: bool = False
        self._change_state = 0  # 0-没有barOpen,barClose,有更新;1-barClose;2-barOpen;3-barClose/barOpen;4-没有更新
        self._instrument_id: string = ""
        self._interval: int = 0
        self._bar_type: EnumBarType = EnumBarType.minute
        self._handlers: defaultdict = defaultdict(list)  # 暂时先不用，后面有需求再放开，通过线程来处理事件，参考vn.py

        self._lock = threading.RLock()

    def register(self, type: str, handler: HandlerType) -> None:
        """
        Register a new handler function for a specific event type. Every
        function can only be registered once for each event type.
        """
        handler_list = self._handlers[type]
        if handler not in handler_list:
            handler_list.append(handler)

    def unregister(self, type: str, handler: HandlerType) -> None:
        """
        Unregister an existing handler function from event engine.
        """
        handler_list = self._handlers[type]

        if handler in handler_list:
            handler_list.remove(handler)

        if not handler_list:
            self._handlers.pop(type)

    def create_bar_provider_by_date_time(self, instrument_manager, instrument_id, begin_time, end_time, interval, bar_type, *instrument_ids):
        """
        根据时间区间创建K线提供器
        :param instrument_manager:
        :param instrument_id:
        :param begin_time:
        :param end_time:
        :param interval:
        :param bar_type:
        :param instrument_ids:
        """
        self._instrument_id = instrument_id
        self._interval = interval
        self._bar_type = bar_type
        if (bar_type == EnumBarType.second) | (bar_type == EnumBarType.minute) | (bar_type == EnumBarType.hour):
            self._lst_date_time_slices = BaseBarHelper.create_in_day_date_time_slice_by_date_time(instrument_manager, instrument_id, begin_time, end_time, interval, bar_type, *instrument_ids)
        elif bar_type == EnumBarType.day:
            self._lst_date_time_slices = BaseBarHelper.create_out_day_date_time_slice_by_date_time(begin_time, end_time, interval, bar_type)
        else:
            raise Exception("不支持的K线类型")

    def create_bar_provider_by_trading_date(self, instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids):
        """
        根据交易日创建K线提供器
        :param instrument_manager:
        :param instrument_id:
        :param trading_date:
        :param interval:
        :param bar_type:
        :param instrument_ids:
        """
        self._instrument_id = instrument_id
        self._interval = interval
        self._bar_type = bar_type
        self._lst_date_time_slices = BaseBarHelper.create_one_day_date_time_slice(instrument_manager, instrument_id, trading_date, interval, bar_type, *instrument_ids)

    def create_bar_provider_by_date(self, instrument_manager, instrument_id, begin_date, end_date, interval, bar_type, *instrument_ids):
        """
        根据日期区间创建K线提供器
        :param instrument_manager:
        :param instrument_id:
        :param begin_date:
        :param end_date:
        :param interval:
        :param bar_type:
        :param instrument_ids:
        """
        self._instrument_id = instrument_id
        self._interval = interval
        self._bar_type = bar_type
        if (bar_type == EnumBarType.second) | (bar_type == EnumBarType.minute) | (bar_type == EnumBarType.hour):
            self._lst_date_time_slices = BaseBarHelper.create_in_day_date_time_slice_by_date(instrument_manager, instrument_id, begin_date,
                                                                                             end_date, interval,
                                                                                             bar_type, *instrument_ids)
        elif bar_type == EnumBarType.day:
            self._lst_date_time_slices = BaseBarHelper.create_out_day_date_time_slice_by_date(begin_date, end_date, interval, bar_type)
        else:
            raise Exception("不支持的K线类型")

    def create_night_am_pm_bar_provider_by_date(self, instrument_manager, instrument_id, begin_date, end_date, *instrument_ids):
        """
        根据日期区间创建K线提供器,只切分夜盘 上午盘 下午盘
        :param instrument_manager:
        :param instrument_id:
        :param begin_date:
        :param end_date:
        :param instrument_ids:
        """
        self._instrument_id = instrument_id
        self._interval = 1
        self._bar_type = EnumBarType.hour
        self._lst_date_time_slices = BaseBarHelper.create_night_am_pm_date_time_slice_by_date(instrument_manager, instrument_id, begin_date, end_date, *instrument_ids)

    @property
    def is_end(self):
        """
        是否结束
        :return:
        """
        return self._pos_time >= len(self._lst_date_time_slices)

    @property
    def bar_series(self):
        """
        K线
        :return:
        """
        return self._bar_series

    def add_bar(self, new_bar):
        """
        添加bar
        :param new_bar:
        :return:
        """
        if new_bar is None:
            self._change_state = 4
            return

        self._lock.acquire()
        if self.is_end:
            self._change_state = 4
            return
        result = self.__move_to(new_bar.begin_time, new_bar.end_time)
        ir = result[0]
        index = result[1]
        if ir == -1:
            self._change_state = 4
            return
        bar_open_count = 0
        bar_close_count = 0
        while self._pos_time < index:
            if self._pos_bar == self._pos_time:
                pass
            else:
                self._pos_bar += 1

            self._pos_time += 1
        if ir == 0:
            if self._pos_bar == self._pos_time:
                bar = self.bar_series[-1]
                bar.add_bar(new_bar)
            else:
                bar = Bar()
                bar.trading_date = new_bar.trading_date
                bar.open_bar_with_new_bar(new_bar)
                self.bar_series.append(bar)

                bar_open_count += 1

                # OnBarOpened(bar)

                self._pos_bar += 1

                current_date_time_slice = self._lst_date_time_slices[self._pos_time]
                if new_bar.end_time >= current_date_time_slice.end_time:
                    bar_close_count += 1

                    self.bar_series[-1].close_bar(current_date_time_slice.end_time)
                    # (BarSeries.Last)

            if (bar_open_count == 0) & (bar_close_count == 0):
                self._change_state = 0
                return

            if bar_open_count > bar_close_count:
                self._change_state = 2
                return

            if bar_close_count > bar_open_count:
                self._change_state = 1
                return

            self._change_state = 3
        self._lock.release()

    def add_tick(self, tick):
        """
        添加Tick
        :param tick:
        :return:
        """
        if tick is None:
            self._change_state = 4
            return

        turnover = 0
        volume = 0
        self._trading_date = tick.trading_date

        self._lock.acquire()
        # 第一个tick
        if self._last_tick is not None:
            turnover = self._last_tick.turnover
            volume = self._last_tick.volume
        else:
            if self._enable_live:
                self._diff_time_span = tick.date_time - datetime.datetime.now()
            for bar in self.bar_series:
                if bar.trading_date == self._trading_date:
                    turnover = turnover + bar.turnover
                    volume = volume + bar.volume

        if self.is_end:
            self._change_state = 4
            return

        result = self.__move_to(tick.date_time, tick.date_time)  # 自动判断是否切换下个区间
        ir = result[0]
        index = result[1]
        if ir == -1:
            self._change_state = 4
            return

        self._last_tick = tick

        bar_open_count = 0
        bar_close_count = 0
        while self._pos_time < index:
            current_t = self._lst_date_time_slices[self._pos_time]

            if self._pos_bar == self._pos_time:
                bar_close_count += 1

                self.bar_series[-1].close_bar(current_t.end_time)
                # OnBarClosed(BarSeries.Last)
            else:
                # 20150628 去掉空Bar
                if self._is_supplement_blank_bar:
                    bar_open_count += 1
                    bar = Bar()
                    bar.begin_time = current_t.begin_time
                    if self._pos_bar >= 0:
                        bar.close = bar.pre_close = bar.open = bar.high = bar.low = self.bar_series[-1].close
                        bar.open_interest = self.bar_series[-1].open_interest
                    elif self._last_tick is not None:
                        if self._last_tick.pre_settlement_price > 0:
                            bar.close = bar.pre_close = bar.open = bar.high = bar.low = self._last_tick.pre_settlement_price
                        else:
                            bar.close = bar.pre_close = bar.open = bar.high = bar.low = self._last_tick.pre_close_price

                        bar.open_interest = self._last_tick.pre_open_interest

                    bar.trading_date = tick.trading_day
                    self.bar_series.append(bar)

                    bar_close_count += 1
                    self.bar_series[-1].close_bar(current_t.end_time)
                    # OnBarClosed(BarSeries.Last)
                self._pos_bar += 1

            self._pos_time += 1

        # 有效的
        if ir == 0:
            if self._pos_bar == self._pos_time:
                bar = self.bar_series[-1]
                bar.add_tick(tick)
                bar.turnover = bar.turnover + tick.turnover - turnover
                bar.volume = bar.volume + tick.volume - volume
            else:
                if len(self.bar_series) == 0:
                    last_bar = None
                else:
                    last_bar = self.bar_series[-1]
                bar = Bar()
                bar.trading_date = self._trading_date
                bar.open_bar(self._lst_date_time_slices[self._pos_time].begin_time, tick, last_bar)
                if last_bar is not None:
                    bar.turnover = tick.turnover - turnover
                    bar.volume = tick.volume - volume
                else:
                    bar.turnover = tick.turnover
                    bar.volume = tick.volume

                self.bar_series.append(bar)

                bar_open_count += 1
                # OnBarOpened(bar)

                self._pos_bar += 1

        if (bar_open_count == 0) & (bar_close_count == 0):
            self._change_state = 0
            return

        if bar_open_count > bar_close_count:
            self._change_state = 2
            return

        if bar_close_count > bar_open_count:
            self._change_state = 1
            return

        self._change_state = 3
        self._lock.release()

    def __move_to(self, begin_time, end_time):
        index = self._pos_time

        # region -1 无效
        current_date_time_slice = self._lst_date_time_slices[self._pos_time]
        if begin_time < current_date_time_slice.begin_time:
            return -1, index
        # endregion

        # region 1 下一个
        while index < len(self._lst_date_time_slices) - 1:
            date_time_slice = self._lst_date_time_slices[index]
            next_date_time_slice = self._lst_date_time_slices[index + 1]

            ir = self.__is_current_date_time_slice(date_time_slice, next_date_time_slice, begin_time, end_time)
            if ir == 0:
                return 0, index
            # 处理盘中休息的异常数据(一般不会出现这种情况)
            if ir == 2:
                index += 1
                return 0, index
            index += 1

        # 收盘以后的Tick处理
        bmy = self.__is_current_date_time_slice(self._lst_date_time_slices[index], None, begin_time, end_time)
        if bmy == 0:
            return 0, index

        index += 1

        # endregion

        return 1, index

    @staticmethod
    def __is_current_date_time_slice(current_date_time_slice, next_date_time_slice, begin_time, end_time):
        if next_date_time_slice is not None:
            if current_date_time_slice.end_time < next_date_time_slice.begin_time:
                if (begin_time > current_date_time_slice.end_time) & (begin_time < next_date_time_slice.begin_time):
                    ta = begin_time - current_date_time_slice.end_time
                    tb = next_date_time_slice.begin_time - begin_time
                    if tb.total_seconds() < ta.total_seconds():
                        return 2
                    return 0

            if (current_date_time_slice.begin_time <= begin_time) & (begin_time < next_date_time_slice.begin_time):
                return 0
        else:
            # 收盘5秒内的tick都计算
            if (current_date_time_slice.begin_time <= begin_time) & (end_time < (current_date_time_slice.end_time + datetime.timedelta(seconds=59))):
                return 0
        return 1

    def _process(self, event: QiEvent) -> None:
        """
        First ditribute event to those handlers registered listening
        to this type.
        """
        if event.type in self._handlers:
            [handler(event) for handler in self._handlers[event.type]]


from enum import Enum, unique


@unique
class EnumOrderPriceType(Enum):
    市价 = '1'
    限价 = '2'
    最优价 = '3'
    最新价 = '4'
    最新价浮动上浮1个ticks = '5'
    最新价浮动上浮2个ticks = '6'
    最新价浮动上浮3个ticks = '7'
    卖一价 = '8'
    卖一价浮动上浮1个ticks = '9'
    卖一价浮动上浮2个ticks = 'A'
    卖一价浮动上浮3个ticks = 'B'
    买一价 = 'C'
    买一价浮动上浮1个ticks = 'D'
    买一价浮动上浮2个ticks = 'E'
    买一价浮动上浮3个ticks = 'F'
    五档价 = 'G'




import datetime

from SapphireQuant.Core.Enum.EnumExchange import EnumExchange


class Exchange:
    """
    交易市场
    """
    def __init__(self):
        self._exchange_id = ""
        self._exchange_name = ""
        self._close_time = None
        self._open_time = None

        self._base_time = None
        self._diff_time = None
        self._all_slice = []

    @property
    def exchange_id(self):
        return self._exchange_id

    @exchange_id.setter
    def exchange_id(self, value):
        self._exchange_id = value

    @property
    def exchange_name(self):
        return self._exchange_name

    @exchange_name.setter
    def exchange_name(self, value):
        self._exchange_name = value

    @property
    def close_time(self):
        return self._close_time

    @close_time.setter
    def close_time(self, value):
        self._close_time = value

    @property
    def open_time(self):
        return self._open_time

    @open_time.setter
    def open_time(self, value):
        self._open_time = value

    @property
    def time_now(self):
        if self._base_time == datetime.datetime:
            return datetime.datetime.now() + self._diff_time
        return self._base_time

    @property
    def base_time(self):
        return self._base_time

    @base_time.setter
    def base_time(self, value):
        self._base_time = value

    @property
    def diff_time(self):
        return self._diff_time

    @diff_time.setter
    def diff_time(self, value):
        self._diff_time = value

    @property
    def all_slice(self):
        return self._all_slice

    @all_slice.setter
    def all_slice(self, value):
        self._all_slice = value

    @property
    def exchange_enum(self):
        if self._exchange_id == "CFFE":
            return EnumExchange.中金所
        elif self._exchange_id == "DCE":
            return EnumExchange.大商所
        elif self._exchange_id == "SHFE":
            return EnumExchange.上期所
        elif self._exchange_id == "CZCE":
            return EnumExchange.郑商所
        elif self._exchange_id == "SH":
            return EnumExchange.上证所
        elif self._exchange_id == "SZ":
            return EnumExchange.深交所
        return EnumExchange.未知

    def to_string(self):
        pass



from setuptools import setup, find_packages

setup(
    name='SapphireQuant',
    version='1.1.5',
    description=(
        'Python Quantitative Investment@2021'
    ),
    # long_description=open('README.rst').read_trading_days(),
    author='CarlSnow',
    author_email='carl.snow.china@gmail.com',
    maintainer='CarlSnow',
    maintainer_email='carl.snow.china@gmail.com',
    license='Apache License',
    packages=find_packages(),
    # py_modules=['Simulate'],
    # package_data={
    #     # If any package contains *.txt or *.rst files, include them:
    #     "Profiles": ["*.tc", "*.xml"],
    #     # And include any *.msg files found in the "hello" package, too:
    #     # "hello": ["*.msg"],
    # },
    platforms=["all"],
    include_package_data=True,
    # exclude_package_data={'': ['setup.py']},
    # zip_safe = False
    url='https://github.com/CarlSnow/SapphireQuant',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries'
    ], install_requires=['pandas', 'openpyxl', 'numpy', 'matplotlib']
    #  install_requires=['pandas']
    # install_requires=['dateutil']
    # install_requires=[
    #     'Twisted>=13.1.0',
    #     'w3lib>=1.17.0',
    #     'queuelib',
    #     'lxml',
    #     'pyOpenSSL',
    #     'cssselect>=0.9',
    #     'six>=1.5.2',
    #     'parsel>=1.1',
    #     'PyDispatcher>=2.0.5',
    #     'service_identity',
    # ]
)



import abc

from SapphireQuant.Core.Enum import EnumOrderStatus
from SapphireQuant.Core.Enum.EnumMarket import EnumMarket

from SapphireQuant.Core.Futures import FutureTrade


class BaseMatchMachine:
    """
    基础撮合机
    """

    def __init__(self):
        self.order_map = {}
        self.match_serial = 0

    def get_can_cancel_orders(self):
        """
        获取可撤的委托单
        :return:
        """
        order_series = []
        for orders in self.order_map.values():
            for order in orders:
                if order.order_status.can_cancel():
                    order_series.append(order)

        return order_series

    def enter(self, order):
        """
        报单
        :param order:
        """
        instrument_id = order.instrument_id
        if instrument_id in self.order_map.keys():
            self.order_map[instrument_id].append(order)
        else:
            self.order_map[instrument_id] = []
            self.order_map[instrument_id].append(order)

    @abc.abstractmethod
    def cancel(self, order):
        """
        撤单
        :param order:
        :return:
        """
        msg = ''
        return False, msg

    @abc.abstractmethod
    def match_by_bar(self, bar):
        """
        撮合
        :param bar:
        :return:
        """
        lst_tuples = []
        if not self.check_bar(bar):
            return lst_tuples

        if len(self.order_map) <= 0:
            return lst_tuples

        if bar.instrument_id not in self.order_map.keys():
            return lst_tuples

        orders = self.order_map[bar.instrument_id]
        for i in range(len(orders)):
            order_item = orders[i]
            match_price = order_item.order_price
            match_volume = order_item.order_volume
            result = self.can_match_by_bar(bar, order_item)
            can_order = result[0]
            if can_order:
                match_item = FutureTrade()
                match_item.task_guid = order_item.task_guid
                # match_item.project_id
                match_item.product_code = order_item.product_code
                match_item.strategy_id = order_item.strategy_id
                match_item.local_id = order_item.local_id

                match_item.order_guid = order_item.guid
                match_item.order_sys_id = order_item.order_sys_id
                match_item.instrument_id = order_item.instrument_id
                match_item.market = EnumMarket.期货
                match_item.instrument_name = order_item.instrument_name
                match_item.exchange_id = order_item.exchange_id
                match_item.trade_volume = order_item.order_volume
                match_item.trade_price = match_price
                match_item.trade_time = bar.end_time
                match_item.direction = order_item.direction
                match_item.open_close = order_item.open_close
                match_item.hedge_flag = order_item.hedge_flag
                self.match_serial += 1
                match_item.trade_sys_id = '{0}'.format(self.match_serial)
                match_item.trading_date = order_item.trading_date

                order_item.volume_traded = match_item.trade_volume
                order_item.match_amount += match_item.trade_price * match_item.trade_volume
                if order_item.volume_traded == order_item.order_volume:
                    order_item.order_status = EnumOrderStatus.全部成交
                else:
                    order_item.order_status = EnumOrderStatus.部分成交
                if order_item.order_volume == match_volume:
                    order_item.remark = '完全成交:成交量{0},剩余{1},价格={2}元@{3}'.format(match_volume, order_item.order_volume - match_volume, match_price,
                                                                               bar.end_time.strftime('%Y/%m%d %H:%M:%S.%f'))
                else:
                    order_item.remark = '部分成交:成交量{0},剩余{1},价格={2}元@{3}'.format(match_volume, order_item.order_volume - match_volume, match_price,
                                                                               bar.end_time.strftime('%Y/%m%d %H:%M:%S.%f'))

                lst_tuples.append([order_item, match_item])
                if order_item.order_volume - order_item.volume_traded <= 0:
                    del orders[i]
                    i -= 1

        return lst_tuples

    @abc.abstractmethod
    def match_by_tick(self, tick):
        """
        撮合
        :param tick:
        :return:
        """
        lst_tuples = []
        if not self.check_tick(tick):
            return lst_tuples

        if len(self.order_map) <= 0:
            return lst_tuples

        if tick.instrument_id not in self.order_map.keys():
            return lst_tuples

        orders = self.order_map[tick.instrument_id]
        for i in range(len(orders)):
            order_item = orders[i]
            match_price = order_item.order_price
            match_volume = order_item.order_volume
            result = self.can_match_by_tick(tick, order_item)
            can_order = result[0]
            if can_order:
                match_item = FutureTrade()
                match_item.task_guid = order_item.task_guid
                # match_item.project_id
                match_item.product_code = order_item.product_code
                match_item.strategy_id = order_item.strategy_id
                match_item.local_id = order_item.local_id

                match_item.order_guid = order_item.guid
                match_item.order_sys_id = order_item.order_sys_id
                match_item.instrument_id = order_item.instrument_id
                match_item.market = EnumMarket.期货
                match_item.instrument_name = order_item.instrument_name
                match_item.exchange_id = order_item.exchange_id
                match_item.trade_volume = order_item.order_volume
                match_item.trade_price = match_price
                match_item.trade_time = tick.date_time
                match_item.direction = order_item.direction
                match_item.open_close = order_item.open_close
                match_item.hedge_flag = order_item.hedge_flag
                self.match_serial += 1
                match_item.trade_sys_id = '{0}'.format(self.match_serial)
                match_item.trading_date = order_item.trading_date

                order_item.volume_traded = match_item.trade_volume
                order_item.match_amount += match_item.trade_price * match_item.trade_volume
                if order_item.volume_traded == order_item.order_volume:
                    order_item.order_status = EnumOrderStatus.全部成交
                else:
                    order_item.order_status = EnumOrderStatus.部分成交
                if order_item.order_volume == match_volume:
                    order_item.remark = '完全成交:成交量{0},剩余{1},价格={2}元@{3}'.format(match_volume, order_item.order_volume - match_volume, match_price,
                                                                               tick.date_time.strftime('%Y/%m%d %H:%M:%S.%f'))
                else:
                    order_item.remark = '部分成交:成交量{0},剩余{1},价格={2}元@{3}'.format(match_volume, order_item.order_volume - match_volume, match_price,
                                                                               tick.date_time.strftime('%Y/%m%d %H:%M:%S.%f'))

                lst_tuples.append([order_item, match_item])
                if order_item.order_volume - order_item.volume_traded <= 0:
                    del orders[i]
                    i -= 1

        return lst_tuples

    @abc.abstractmethod
    def check_bar(self, bar):
        """
        检查
        :param bar:
        :return:
        """
        return True

    @abc.abstractmethod
    def check_tick(self, tick):
        """
        检查
        :param tick:
        :return:
        """
        return True

    @abc.abstractmethod
    def can_match_by_bar(self, bar, order):
        """
        是否可以撮合
        :param bar:
        :param order:
        :return:
        """
        match_price = None
        match_volume = None
        return False, match_price, match_volume

    @abc.abstractmethod
    def can_match_by_tick(self, tick, order):
        """
        是否可以撮合
        :param tick:
        :param order:
        :return:
        """
        match_price = None
        match_volume = None
        return False, match_price, match_volume



import datetime
import os

import numpy as np
import pandas as pd

from SapphireQuant.DataStorage.Cache.FutureCache import EnumLiquidationType
from SapphireQuant.Lib.FileToolBox import FileToolBox

from SapphireQuant.Config.TradingDay import TradingDayHelper
from SapphireQuant.Core.Enum import EnumOpenClose, EnumPositionDirection, EnumBuySell, EnumMarket
from SapphireQuant.Core.Futures import FutureTrade
from SapphireQuant.Strategy.BaseStrategy import BaseStrategy


class StrategyEngine:
    """
    策略引擎
    """
    @staticmethod
    def simulate_by_trade(**kwargs):
        """
        利用Trade记录进行复盘
        :param kwargs:
                record_dir:输出记录的目录
                df_trade:[ProductCode, StrategyId, InstrumentId, Direction, OpenClose, Price, Volume, TradeTime, TradingDate, AccountId]
                # df_position:[ProductCode, StrategyId, InstrumentId, Direction, Volume, TradingDate, AccountId]
        :return:
        """
        df_trade = kwargs['df_trade']
        record_dir = kwargs['record_dir']
        df_trade['TradeTime'] = pd.to_datetime(df_trade['TradeTime'])
        df_trade['TradingDate'] = pd.to_datetime(df_trade['TradingDate'])
        df_all_asset = None
        df_all_summary = None
        for product_code, df_product_trade in df_trade.groupby(['ProductCode']):
            df_product_asset = None
            df_product_summary = None
            for strategy_id, df_strategy_trade in df_product_trade.groupby(['StrategyId']):
                begin_date = df_strategy_trade['TradingDate'].min()
                end_date = df_strategy_trade['TradingDate'].max()
                kwargs['drive_code'] = ''
                kwargs['subscribe_futures'] = ''
                kwargs['drive_bar_interval'] = None
                kwargs['drive_bar_type'] = None
                kwargs['start_simulate_date'] = begin_date
                kwargs['end_simulate_date'] = end_date
                kwargs['product_code'] = product_code
                kwargs['strategy_id'] = strategy_id
                kwargs['liquidation_type'] = EnumLiquidationType.收盘价
                strategy = BaseStrategy(**kwargs)
                print('{0}-{1}:{2}-{3}'.format(product_code, strategy_id, begin_date, end_date))
                lst_trading_days = TradingDayHelper.get_trading_days(begin_date, end_date)
                for trading_day in lst_trading_days:
                    strategy.day_begin(trading_day, trading_day)
                    df_today_trade = df_strategy_trade[df_strategy_trade['TradingDate'] == trading_day]
                    if len(df_today_trade) > 0:
                        df_today_trade = df_today_trade.sort_values(by="TradeTime", ascending=True).copy()
                        for index, row in df_today_trade.iterrows():
                            trade = FutureTrade()
                            trade.product_code = row['ProductCode']
                            trade.strategy_id = row['StrategyId']
                            trade.account_id = row['AccountId']
                            trade.instrument_id = row['InstrumentId']
                            trade.direction = row['Direction']
                            trade.open_close = row['OpenClose']
                            trade.trade_volume = row['Volume']
                            trade.trade_price = row['Price']
                            trade.trading_date = row['TradingDate']
                            trade.trade_time = row['TradeTime']
                            print('成交:{0}'.format(trade.to_string()))
                            strategy.future_cache.on_trade(trade)
                    for position in strategy.future_cache.get_future_positions():
                        # 主力合约不处理交割问题
                        if '9999' in position.instrument_id:
                            continue
                        if position.volume > 0:
                            if 'IC' in position.instrument_id or 'IF' in position.instrument_id or 'IH' in position.instrument_id:
                                year = int(position.instrument_id[2:4]) + 2000
                                month = int(position.instrument_id[4:6])
                                month_first_day = datetime.datetime(year, month, 1)
                                week_day = month_first_day.weekday()
                                if week_day <= 4:
                                    expire_date = month_first_day + datetime.timedelta(days=2 * 7 + 4 - week_day)
                                else:
                                    expire_date = month_first_day + datetime.timedelta(days=2 * 7 + 6 - week_day + 5)
                                expire_date = TradingDayHelper.get_first_trading_day(expire_date)
                            else:
                                future_info = kwargs['instrument_manager'][position.instrument_id]
                                expire_date = future_info.expire_date

                            if expire_date == trading_day:
                                print('到了交割日，还未交割，自动执行交割{0}'.format(expire_date))
                                trade = FutureTrade()
                                trade.product_code = position.product_code
                                trade.strategy_id = position.strategy_id
                                trade.account_id = position.account_id
                                trade.instrument_id = position.instrument_id
                                trade.trade_volume = position.volume
                                trade.trading_date = position.trading_date
                                trade.trade_time = position.trading_date + datetime.timedelta(hours=15)
                                trade.open_close = EnumOpenClose.平仓
                                if position.direction == EnumPositionDirection.多头:
                                    trade.direction = EnumBuySell.卖出
                                else:
                                    trade.direction = EnumBuySell.买入

                                day_bar_series = kwargs['data_handler'].load_day_bar_series(EnumMarket.期货, position.instrument_id, position.trading_date)
                                trade.trade_price = day_bar_series[0].平仓
                                print('交割单:{0}'.format(trade.to_string()))
                                strategy.future_cache.on_trade(trade)
                    strategy.day_end()
                strategy.generate_strategy_report()
                strategy.save_strategy(os.path.join(record_dir, '{0}.xlsx'.format(product_code)), os.path.join(record_dir, '{0}.png'.format(product_code)))
                df_temp = strategy.df_asset[['日期', '累计盈亏']]
                df_temp.rename(columns={'累计盈亏': '{0}_{1}'.format(product_code, strategy_id)}, inplace=True)
                if df_product_asset is None:
                    df_product_asset = df_temp
                else:
                    df_product_asset = df_product_asset.merge(df_temp, how='outer', on=['日期'])
                if df_product_summary is None:
                    df_product_summary = strategy.df_summary
                else:
                    df_product_summary = df_product_summary.append(strategy.df_summary, ignore_index=True)

            record_path = '{0}/{1}/产品记录明细.xlsx'.format(record_dir, product_code)
            if df_product_asset is None:
                continue
            df_product_asset = df_product_asset.replace(np.nan, 0)
            df_product_asset_diff = df_product_asset.diff()
            df_product_asset_diff = df_product_asset_diff.replace(np.nan, 0)
            df_product_asset_diff['日期'] = df_product_asset['日期']

            dic_product = {
                'Asset': df_product_asset,
                'Asset_In_Day': df_product_asset_diff,
                'Summary': df_product_summary
            }

            FileToolBox.save_excel(record_path, **dic_product)

            if df_all_asset is None:
                df_all_asset = df_product_asset
            else:
                df_all_asset = df_all_asset.merge(df_product_asset, how='outer', on=['日期'])
            if df_all_summary is None:
                df_all_summary = df_product_summary
            else:
                df_all_summary = df_all_summary.append(df_product_summary, ignore_index=True)

        record_path = '{0}/账号记录明细.xlsx'.format(record_dir)
        if df_all_asset is None:
            return
        df_all_asset = df_all_asset.replace(np.nan, 0)
        df_all_asset_diff = df_all_asset.diff()
        df_all_asset_diff['日期'] = df_all_asset['日期']
        df_all_asset_diff = df_all_asset_diff.replace(np.nan, 0)

        dic_all = {
            'Asset': df_all_asset,
            'Asset_In_Day': df_all_asset_diff,
            'Summary': df_all_summary
        }

        FileToolBox.save_excel(record_path, **dic_all)



import datetime


class QiDataDirectory:
    """
    目录配置
    """
    def __init__(self):
        self.__trading_day = datetime.date
        self.__future_tick = ""
        self.__future_tick_cache = ""
        self.__future_min = ""
        self.__future_day = ""

    @property
    def trading_day(self):
        """

        :return:
        """
        return self.__trading_day

    @trading_day.setter
    def trading_day(self, value):
        self.__trading_day = value

    @property
    def future_tick(self):
        return self.__future_tick

    @future_tick.setter
    def future_tick(self, value):
        self.__future_tick = value

    @property
    def future_tick_cache(self):
        return self.__future_tick_cache

    @future_tick_cache.setter
    def future_tick_cache(self, value):
        self.__future_tick_cache = value

    @property
    def future_min(self):
        return self.__future_min

    @future_min.setter
    def future_min(self, value):
        self.__future_min = value

    @property
    def future_day(self):
        return self.__future_day

    @future_day.setter
    def future_day(self, value):
        self.__future_day = value



from enum import Enum, unique


@unique
class EnumOrderRejectReason(Enum):
    每秒请求超许可数 = 0  # 已撤单报单被拒绝综合交易平台：交易所每秒发送请求数超过许可数
    价格非最小单位倍数 = 1  # 已撤单报单被拒绝价格非最小单位的倍数
    价格高于涨停板 = 2   # 已撤单报单被拒绝价格涨破涨停板
    价格低于跌停板 = 3 # 已撤单报单被拒绝价格跌破跌停板
    资金不足 = 4   # 综合交易平台：资金不足
    平今仓位不足 = 5   # 综合交易平台：平今仓位不足
    平昨仓位不足 = 6    # 综合交易平台：平昨仓位不足
    不支持报单类型= 7  # 已撤单报单被拒绝不被支持的报单类型
    不支持套利类型报单= 8  # 该交易所不支持套利类型报单
    保值额度不足= 9  # 已撤单报单被拒绝保值额度不足
    网络连接失败= 10  # 网络连接失败
    延迟开仓被Qi拒绝= 11  # 委托数超过流量限制，MagicQuant拒绝延迟发送开仓单
    Qi暂停开仓= 12 # MagicQuant暂停发送开仓单XX秒
    报单数量错误= 13
    提交委托失败= 14  # 发送委托失败
    未知= 15 # 未知


class FutureTradingFrame:
    """
    期货交易时间框架
    """
    def __init__(self):
        self._exchanges = []

    @property
    def exchanges(self):
        """

        :return:
        """
        return self._exchanges

    @exchanges.setter
    def exchanges(self, value):
        self._exchanges = value



from SapphireQuant.Config.TradingFrame.LimitValue import LimitValue


class FutureProductTradingFrame:
    """
    产品TradingFrame
    """
    def __init__(self):
        self._id = ""
        self._name = ""
        self._trading_frames = []
        self._limit_value = LimitValue()

    @property
    def id(self):
        """

        :return:
        """
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def name(self):
        """

        :return:
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def trading_frames(self):
        """

        :return:
        """
        return self._trading_frames

    @trading_frames.setter
    def trading_frames(self, value):
        self._trading_frames = value

    @property
    def limit_value(self):
        """

        :return:
        """
        return self._limit_value

    @limit_value.setter
    def limit_value(self, value):
        self._limit_value = value
